/*

 shint_c

 Internal Commands

 Date		Author		Comments
 ----		------		--------
 10/10/1991	P.J. Taylor	Original.
 26/02/1998	A.D. Ives	Reformatted code (put ifs 'right' way round).
				Fixed !arg (should have been !*arg) for many
				of the commands.
				ANSI Function headers adopted.
				kill() now uses sscanf() to get Job and Tag.
 01/03/1998	A.D.Ives	SwitchEcho() added to support ECHO command.
				SetPrompt() added to support PROMPT command.
 02/03/1998	A.D.Ives	fexec() added. Calls fi2exec() (imported
				from Jonathon Hudson's QasQade) to try and
				execute a data file. This is called at the
				bottom of the tests in sh_c, when no other
				internal command was matched.
 03/03/1998	A.D.Ives	CD command enchanced to support MS DOS style
				navigation shorthand:-

				  \	 = Root directory
				  ..	 = Parent directory
				  ..\Dir = New directory at same level
 04/03/1998	A.D.Ives	Enhanced navigation also added to CPD and
				CDD commands (although unlikely to be used).
 06/03/1998	A.D.Ives	md() function added.
				Changed format of error messages.

 1.04
 ----
 07/03/1998	A.D.Ives	Added WaitKey() to support PAUSE command.
				Added IfTest() to support simple tests.
				Added FTest() to check for the existence of
				a file.

 1.05
 ----
 09/03/1998	A.D.Ives	Rationalised cd() so that it takes an
				additional integer parameter that
				specifies which directory is to be
				changed.
				Added SetOSDir() function to set the
				OS dir equal to our local directory.
				Modified cd() so that it detects special
				prefix '@' and synchronises the QDOS/SMSQ
				directory if present.
				Added let() to set a local environment
				variable.
				Added print() to print out a string.
				Added opt() to support new command OPT to
				set/unset Shell options.

 1.06
 ----
 10/03/1998	A.D.Ives	UpDir(), RootDir(), DirChange() and SetOSDir()
				moved into misc_c.
				FTest() now calls NavName() to handle the
				special directory navigation prefixes.
				fexec() also calls NavName().
				cd(), md(), fexec() all translate forward
				slashes into backwards slashes in their
				argument string.
				cd() and md() interpret the new Shell options
				+u (shift to upper case) and +l (shift to
				lower case).
				pause() stores the key used to exit in
				gbLastKey to be accessed as a new internal
				variable.
  11/03/1998			globals gbResult and gbRetCode replaced by
				statics in new module suppt. Values now
				accessed via the Get/Set function SI().
				New string comparison operators:-
				  begins  ends
  12/03/1998			winsize(), winpos(), and border()
				rationalised to use new SetCon() routine.

  1.08
  ----
  27/03/1998			docmd() added.
  03/04/1998			Extended DO command to support multiple
				ranges. Did the same for FOR. Both commands
				now expand parameters in their arguments.

  1.09
  ----
  17/04/1998			DTest() added.
				Added support for $/ in FOR command (gets
				filename only)
				alias() added.

  26/04/1998			cd() supports "=" as a parameter, to set
				equal to current OS dir.

*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <qdos.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <debug.h>
#include "sh.h"
#include "shint.h"
#include "suppt.h"
#include "libpjt.h"

extern int argCount ;		/* Count of command line arguments	*/
extern char **argStrings ;	/* Array of command line arguments	*/

extern char *_VENV ;		/* Pointer to the local environment	*/

void cls() {

  /* CLS - Clear the Console window */

  sd_clear(fgetchid(stdout), (timeout_t)(-1)) ;
}

void ink(char *arg) {

  /* INK Colour - Set Console ink colour */

  int n ;
  if ( arg != NULL && isdigit(*arg) )
    n= atoi(arg) ;
  else {
    fprintf(stderr, "INK: no parameter supplied\n") ;
    return ;
  }
  sd_setin(fgetchid(stdout), (timeout_t)(-1), n) ;
}

void paper(char *arg) {

  /* PAPER Colour - Set Console paper colour */

  int n ;
  if ( arg != NULL && isdigit(*arg) )
    n= atoi(arg) ;
  else {
    fprintf(stderr, "PAPER: no parameter supplied\n") ;
    return ;
  }
  sd_setpa(fgetchid(stdout), (timeout_t)(-1), n) ;
  sd_setst(fgetchid(stdout), (timeout_t)(-1), n) ;
}

void winpos(char *arg) {

  /* WINPOS Xpos YPos - Set console window position */

  char def[128] ;

  strcpy(def, "-1 -1 ") ;
  strcat(def, arg) ;
  SetCon(fgetchid(stdout), def) ;
}

void winsize(char *arg) {

  /* WINSIZE XSize YSize - Set Console window size */

  SetCon(fgetchid(stdout), arg) ;
}

void border(char *arg) {

  /* BORDER Width Ink - Set Console border colour and width */

  char def[128] ;

  strcpy(def, "-1 -1 -1 -1 ") ;
  strcat(def, arg) ;
  SetCon(fgetchid(stdout), def) ;
}

void waitabit(char *arg) {

  /* WAIT n - Wait for n seconds */

  int n;
  if ( arg != NULL && isdigit(*arg) )
    n= atoi(arg) ;
  else
    n= 5 ;
  if ( n < 1 )
    n= 1 ;
  if ( errno= mt_susjb(-1L, n * FR_SEC, NULL) )
    qerror( "WAIT: Failed to suspend self" );
}

void kill(char *arg) {

  /* KILL Job Tag - Force Remove a Job */

  int n, job, tag ;
  jobid_t jobid;  /* ID of job to be removed */

  if ( arg == NULL || !*arg ) {
    fprintf(stderr, "KILL: Syntax Error: Parameters missing\n") ;
    return ;
  }

  if ( ( n= sscanf(arg, "%d %d", &job, &tag) ) != 2 ) {
    fprintf(stderr, "KILL: Not enough parameters supplied\n") ;
    return ;
  }
  jobid= ((tag & 0xFFFF) << 16) | (job & 0xFFFF) ;
  if ( jobterm(jobid) )
    qerror2("KILL: Cannot remove Job", arg) ;
}

void cd(char *arg, int which) {

  /* CD {Dir} - Change Directory */

  char Dir[64], osdir[64] ;
  int res, sync= FALSE ;

  if ( strcmp(arg,"=") == 0 )

    /* CD = gets the current OS Dir */

    arg= GetOSDir(osdir, which) ;

  SlashToBackslash(arg) ;
  if ( ShellOpt("+u") )
    strupr(arg) ;
  if ( ShellOpt("+l") )
    strlwr(arg) ;
  if ( *arg == '@' ) {
    sync= TRUE ;
    arg++ ;
  }
  sync|= ShellOpt("+d") ;
  switch( which ) {
    case D_DATAD :
      getcwd(Dir, 64) ;
      break ;
    case D_PROGD :
      getcpd(Dir, 64) ;
      break ;
    case D_DESTD :
      getcdd(Dir, 64) ;
      break ;
  }
  if ( arg == NULL || !*arg ) {
    if (sync)
      SetOSDir(which) ;
    else
      printf ("%s\n", Dir) ;
  }
  else {
    switch( which ) {
      case D_DATAD :
	res= chdir(DirChange(Dir, arg)) ;
	break ;
      case D_PROGD :
	res= chpdir(DirChange(Dir, arg)) ;
	break ;
      case D_DESTD :
	res= chddir(DirChange(Dir, arg)) ;
	break ;
    }
    if ( res )
      qerror2("Cannot change to directory", arg) ;
    else
      if ( sync )
	SetOSDir(which) ;
  }
}

void SwitchEcho(char *arg, int *Echo) {

  /* ECHO ON | OFF - Enable/Disable command echo */

  *Echo= ( stricmp(arg, "on") == 0 ) ;

}

void SetPrompt(char *arg, char *Prompt) {

  /* PROMPT String - Set the prompt */

  strcpy(Prompt, arg) ;

}

int fexec(char *Name) {

  /* Try and "execute" a file using FileInfo II */

  int Result, ChanOpened ;
  chanid_t Chan ;
  char Dir[64], FullName[128] ;

  /*

    fi2exec() needs a CONsole channel - see if we've got one.
    If not, open one for the duration.

  */

  SlashToBackslash(Name) ;
  ChanOpened= FALSE ;
  if ( isatty(fileno(stdin)) )
    Chan= fgetchid(stdin) ;
  else
    if ( isatty(fileno(stdout)) )
      Chan= fgetchid(stdout) ;
    else {
      Chan= io_open("CON", 0) ;
      ChanOpened= TRUE ;
    }

  if ( Name[0] == '\\' || Name[0] == '.' ) {

    /*

      If the name includes a directory navigation prefix process
      it at once, and don't try any further defaulting.

    */

    strcpy(FullName, Name) ;	/* NavName() needs more space */
    Result= fi2exec(NavName(FullName), Chan) ;
  }
  else {
    Result= fi2exec(Name, Chan) ;
    if (Result < 0 ) {
      getcwd(Dir, 64) ;
      strcpy(FullName, Dir) ;
      strcat(FullName, Name) ;
      Result= fi2exec(FullName, Chan) ;
    }
  }

  /* Close any temporary channel that we opened */

  if ( ChanOpened )
    io_close(Chan) ;

  return (Result >= 0) ;
}

void md(char *arg) {

  /* Make a Directory */

  char Dir[64], NewDir[64] ;

  SlashToBackslash(arg) ;
  if ( ShellOpt("+u") )
    strupr(arg) ;
  if ( ShellOpt("+l") )
    strlwr(arg) ;
  if ( (strcmp(arg, "\\") == 0) || (strcmp(arg, "..") == 0) )
    fprintf(stderr, "MD: Cannot make directory %s\n", arg) ;
  else {
    getcwd(Dir, 64) ;
    strcpy(NewDir, DirChange(Dir, arg)) ;
    if ( mkdir(NewDir) != 0 )
      fprintf(stderr, "MD: Cannot make directory %s\n", NewDir) ;
  }
}

void WaitKey(char *msg) {

  /* Pause and wait for the user to hit a key */

  FILE *fp ;
  chanid_t Chan ;
  int ChanOpened ;
  char LastKey ;

  ChanOpened= FALSE ;
  if ( isatty(fileno(stdin)) ) {
    Chan= fgetchid(stdin) ;
    fp= stdin ;
  }
  else {
    if ( isatty(fileno(stdout)) ) {
      Chan= fgetchid(stdout) ;
      fp= stdout ;
    }
    else {
      fp= fopen("CON", "rw") ;
      Chan= fgetchid(fp) ;
      ChanOpened= TRUE ;
    }
  }

  if ( msg != NULL && msg[0] )
    io_sstrg(Chan, -1, msg, (short)(strlen(msg))) ;
  sd_cure(Chan, -1) ;
  io_fbyte(Chan, -1, &LastKey) ;
  sd_curs(Chan, -1) ;
  io_sbyte(Chan, -1, 10) ;

  if ( ChanOpened )
    fclose(fp) ;

  SI(SI_LASTKEY, &LastKey) ;
}

int IfTest(char *arg) {

  /* Perform an IF test, and return TRUE if the test succeeds */

  char Left[128] ,	/* Left operand 	*/
       Op[128] ,	/* Operator		*/
       Right[128] ,	/* Right operand	*/
       *ptr ;
  int  Result, i ;

  i= strlen(arg) ;
  if ( i == 0 )
    return FALSE ;
  ptr= arg + i - 1 ;

  /* Look for the last non-blank character */

  while ( ptr > arg && ( *ptr == ' ' || *ptr == '\t' ) )
    ptr-- ;

  /* Where is the first space? */

  i= strpos(arg, ' ') ;
  if ( ( i != -1 ) && ( arg + i > ptr ) )
    i= -1 ;	/* After the last non-blank, so it's a trailer */

  if ( i == -1 ) {

    /* Default is a simple IF True/False test */

    if ( strlen(arg) > 1 && *arg == '!' ) {

      /* Logical NOT */

      strcpy(Left, arg + 1) ;
      strcpy(Op, "=") ;
    }
    else {
      strcpy(Left, arg) ;
      strcpy(Op, "<>") ;
    }
    strcpy(Right, "0") ;
  }
  else
    sscanf(arg, "%s %s %s", Left, Op, Right) ;

  if ( strcmp(Op, "==") == 0 )		/* String equality	*/
    Result= (strcmp(Left, Right) == 0) ;
  else
  if ( strcmp(Op, "!=") == 0 )		/* String inequality	*/
    Result= (strcmp(Left, Right) != 0) ;
  else
  if ( strcmpi(Op, "in") == 0 ) 	/* Instr		*/
    Result= (strstr(Right, Left) != NULL) ;
  else
  if ( strcmpi(Op, "begins") == 0 )	/* Begins with		*/
    Result= (strstr(Left, Right) == Left) ;
  else
  if ( strcmpi(Op, "ends") == 0 )	/* Ends with		*/
    Result= (strrstr(Left, Right) == (Left + strlen(Left) - strlen(Right))) ;
  else
  if ( strcmpi(Op, "matches") == 0 )	/* Wildcard match	*/
    Result= fnmatch(Left, Right) ;
  else
  if ( strcmp(Op, "=") == 0)		/* Numeric equality	*/
    Result= (atof(Left) == atof(Right)) ;
  else
  if ( strcmp(Op, "<>") == 0)		/* Numeric inequality	*/
    Result= (atof(Left) != atof(Right)) ;
  else
  if ( strcmp(Op, ">") == 0)
    Result= (atof(Left) > atof(Right)) ;
  else
  if ( strcmp(Op, "<") == 0)
    Result= (atof(Left) < atof(Right)) ;
  else
  if ( strcmp(Op, ">=") == 0)
    Result= (atof(Left) >= atof(Right)) ;
  else
  if ( strcmp(Op, "<=") == 0)
    Result= (atof(Left) <= atof(Right)) ;

  return Result ;
}

void FTest(char *Name) {

  /* Set result according to whether a file exists or not */

  char LongName[128] ;
  long Result ;

  strcpy(LongName, Name) ;	/* NavName() may need more space */
  Result= ( access(NavName(LongName), 0) == 0 ) ;
  SI(SI_RESULT, &Result) ;
}

void DTest(char *Name) {

  /* Set result according to whether a directory exists or not */

  char LongName[128] ;
  long Result ;

  strcpy(LongName, Name) ;	/* NavName() may need more space */
  Result= IsDir(NavName(LongName)) ;
  SI(SI_RESULT, &Result) ;
}

void let(char *arg) {

  /* Assign a value to a local environment variable */

  if ( arg != NULL )
    putenv(arg) ;
}

void print(char *arg) {

  /* Print a string on stdout */

  printf("%s\n", arg) ;
}

void opt(char *arg) {

  /* Set a Shell option */

  char *ptr, *opts, options[256] ;

  opts= getenv("SHELL_OPTS") ;
  if ( arg == NULL || !arg[0] ) {
    if ( opts != NULL )
      printf("%s\n", opts) ;
  }
  else {
    strcpy(options, "SHELL_OPTS=") ;
    if ( opts != NULL )
      strcat(options, opts) ;
    ptr= strstr(options, arg) ;
    if ( ptr == NULL ) {

      /* Not there, so add it */

      strcat(options, arg) ;
    }
    else {

      /* Already there, so lose it */

      strcpy(ptr, ptr + strlen(arg)) ;
    }
    putenv(options) ;
  }
}

void DoToDir(char *wildcard, char *cmd) {

  /* Do a Shell command line for every matching file */

  DIR_LIST_t *dir, *Ptr ;
  qdirect_t This ;
  char forvar[128], forwc[128], forfn[128], forfnb[128], lbuff[256] ;
  int devend, wc ;

  devend= strpos(wildcard, '_') ;
  if ( devend > 1 && devend < 10 ) {

    /* Find the device */

    strncpy(forvar, wildcard, ++devend) ;
    forvar[devend]= 0 ;
    strcpy(forwc, forvar) ;

    dir= qdir_read(wildcard, "N", 3) ;
    Ptr= dir ;
    while ( Ptr ) {
      This= Ptr->dl_dir ;

      /* Set the FOR variable */

      strcpy(forvar + devend, This.d_name) ;
      SI(SI_FOR, forvar) ;

      /* Get the base filename and store it for retrieval as $/ */

      FileName(forfn, forvar) ;
      SI(SI_FORF, forfn) ;

      /* Set the base filename minus extension for retrieval as $\ */

      wc= strrpos(forfn, '_') ;
      if ( wc == -1 )
	*forfnb= 0 ;
      else {
	strncpy(forfnb, forfn, wc) ;
	forfnb[wc]= 0 ;
      }
      SI(SI_FORFN, forfnb) ;

      /* Set the FOR variable - minus extension  */

      wc= strrpos(This.d_name, '_') ;
      if ( wc == -1 )
	*forwc= 0 ;
      else {
	strncpy(forwc + devend, This.d_name, wc) ;
	forwc[devend + wc]= 0 ;
      }
      SI(SI_FORW, forwc) ;

      /* Do the command */

      strcpy(lbuff, cmd) ;
      Process(lbuff) ;
      Ptr= Ptr->dl_next ;
    }
    qdir_delete(dir) ;
    SI(SI_FOR, "\0") ;
    SI(SI_FORW, "\0") ;
    SI(SI_FORF, "\0") ;
  }
}


void forcmd(char *arg) {

  /* Execute a command FOR every matching file */

  int i ;
  char buff[64] ;

  i= strlen(arg) - 1 ;
  if ( ( *arg == '"' ) && ( *(arg + i) == '"' ) && i > 0 ) {

    /* Remove enclosing quotes */

    arg++ ;
    *(arg + --i)= 0 ;
  }

  i= strpos(arg, ' ') ;
  if ( i == -1 )
    fprintf(stderr, "FOR: Syntax Error\n") ;
  else {

    char save[256], wildcard[256], *cmd ;
    int more, next ;

    arg[i]= 0 ;
    strcpy(wildcard, arg) ;
    more= TRUE ;
    *save= 0 ;

    while (more) {

      if ( *save )		/* Anything left in the buffer? */
	strcpy(wildcard, save) ;	/* Yes, so use it.	       */
      *save= 0 ;
      next= strpos(wildcard, ',') ;	/* Look for the next range	*/
      if ( next > 0 ) {
	strcpy(save, wildcard + next + 1) ;
	wildcard[next]= 0 ;	/* Save next wildcard		*/
      }
      more= (*save) ;		/* Indicate if still more to do */

      /*

	Have to do the parameter substitution here, because it is
	inhibited for the command.

      */

      SubstParams(wildcard, argCount, argStrings) ;

      cmd= arg + i + 1 ;
      if ( !isdirdev(wildcard) ) {
	getcwd(buff, 64) ;
	strcat(buff, wildcard) ;
	strcpy(wildcard, buff) ;
      }
      DoToDir(wildcard, cmd) ;
    } /* while (more) */
  }
}

void docmd(char *arg) {

  /* Do a command for a range of values */

  int i, more, next, start, finish, lvar, num ;
  char buff[20], lbuff[256], save[256] ;

  i= strlen(arg) - 1 ;
  if ( ( *arg == '"' ) && ( *(arg + i) == '"' ) && i > 0 ) {

    /* Remove enclosing quotes */

    arg++ ;
    *(arg + --i)= 0 ;
  }

  i= strpos(arg, ' ') ;
  if ( i == -1 )
    fprintf(stderr, "DO: Syntax Error\n") ;
  else {

    char from[256], to[256], *cmd ;
    int simple ;

    arg[i]= 0 ;
    cmd= arg + ++i ;
    *save= 0 ;
    more= TRUE ;

    while (more) {

      if ( *save )		/* Anything left in the buffer? */
	strcpy(arg, save) ;	/* Yes, so use it.		*/
      *save= 0 ;
      next= strpos(arg, ',') ;	/* Look for the next range	*/
      if ( next > 0 ) {
	strcpy(save, arg + next + 1) ;
	arg[next]= 0 ;		/* Save next range		*/
      }
      more= (*save) ;		/* Indicate if still more to do */

      i= strpos(arg, ':') ;
      if ( i == -1 ) {          /* No range, just a single value */
	strcpy(from, arg) ;
	strcpy(to, arg) ;
      }
      else {
	strncpy(from, arg, i) ;
	from[i]= 0 ;
	strcpy(to, arg + i + 1) ;
      }

      /*

	Have to do the parameter substitution here, because it is
	inhibited for the command.

      */

      SubstParams(from, argCount, argStrings) ;
      SubstParams(to, argCount, argStrings) ;

      if ( isdigit(*to) && isdigit(*from)) {
	num= TRUE ;
	start= atoi(from) ;
	finish= atoi(to) ;
      }
      else {
	num= FALSE ;
	start= *from ;
	finish= *to ;
	buff[1]= 0 ;
      }
      if ( finish >= start ) { ;
	for ( lvar= start ; lvar <= finish ; lvar++ ) {
	  if ( num )
	    sprintf(buff, "%u", lvar) ;
	  else
	    *buff= lvar ;
	  SI(SI_DO, buff) ;
	  strcpy(lbuff, cmd) ;
	  Process(lbuff) ;
	}
      }
      else {
	for ( lvar= start ; lvar >= finish ; lvar-- ) {
	  if ( num )
	    sprintf(buff, "%u", lvar) ;
	  else
	    *buff= lvar ;
	  SI(SI_DO, buff) ;
	  strcpy(lbuff, cmd) ;
	  Process(lbuff) ;
	}
      }
      SI(SI_DO, "\0") ;
    } /* while (more) */
  }
}

void alias(char *arg) {

  /* Set/remove an alias */

  int WhereEq, i ;
  char env[256], wc[6], *ptr= _VENV ;

  if ( strlen(arg) == 0 ) {

    /* No parameters get a list of defined aliases */

    strcpy(wc, "SH_*") ;
    while ( *ptr ) {
      if ( fnmatch(ptr, wc) )
	printf("%s\n", ptr + 3) ;
      ptr+= strlen(ptr) + 1 ;
    }
  }
  else {

    i= strlen(arg) - 1 ;
    if ( ( *arg == '"' ) && ( *(arg + i) == '"' ) && i > 0 ) {

      /* Remove enclosing quotes */

      arg++ ;
      *(arg + --i)= 0 ;
    }

    /* Make sure the alias name is always stored in upper case */

    WhereEq= strpos(arg, '=') ;
    if ( WhereEq == -1 )
      strupr(arg) ;
    else {
      arg[WhereEq]= 0 ;
      strupr(arg) ;
      arg[WhereEq]= '=' ;
    }

    strcpy(env, "SH_") ;
    strcat(env, arg) ;
    putenv(env) ;
  }
}
